@@ -39,6 +39,8 @@
 #if EXCLUDE_GENA == 0
 #ifdef INCLUDE_DEVICE_APIS
 
+#include <assert.h>
+
 #include "gena.h"
 #include "httpparser.h"
 #include "httpreadwrite.h"
@@ -54,6 +56,8 @@
 	#define snprintf _snprintf
 #endif
 
+#define STALE_JOBID (INVALID_JOB_ID -1)
+
 /*!
  * \brief Unregisters a device.
  *
@@ -314,9 +318,6 @@ static void genaNotifyThread(
 	notify_thread_struct *in = (notify_thread_struct *) input;
 	int return_code;
 	struct Handle_Info *handle_info;
-	ThreadPoolJob job;
-
-	memset(&job, 0, sizeof(job));
 
 	/* This should be a HandleLock and not a HandleReadLock otherwise if there
 	 * is a lot of notifications, then multiple threads will acquire a read
@@ -340,22 +341,6 @@ static void genaNotifyThread(
 		HandleUnlock();
 		return;
 	}
-#ifdef UPNP_ENABLE_NOTIFICATION_REORDERING
-	/*If the event is out of order push it back to the job queue */
-	if (in->eventKey != sub->ToSendEventKey) {
-		TPJobInit(&job, (start_routine) genaNotifyThread, input);
-		TPJobSetFreeFunction(&job, (free_function) free_notify_struct);
-		TPJobSetPriority(&job, MED_PRIORITY);
-		/* Sleep a little before creating another thread otherwise if there is
-		 * a lot of notifications to send, the device will take 100% of the CPU
-		 * to create threads and push them back to the job queue. */
-		imillisleep(1);
-		ThreadPoolAdd(&gSendThreadPool, &job, NULL);
-		freeSubscription(&sub_copy);
-		HandleUnlock();
-		return;
-	}
-#endif
 
 	HandleUnlock();
 
@@ -380,6 +365,25 @@ static void genaNotifyThread(
 	if (sub->ToSendEventKey < 0)
 		/* wrap to 1 for overflow */
 		sub->ToSendEventKey = 1;
+
+	/* Remove head of event queue. Possibly activate next */
+	{
+		ListNode *node = ListHead(&sub->outgoing);
+		if (node)
+			ListDelNode(&sub->outgoing, node, 1);
+		if (ListSize(&sub->outgoing) > 0) {
+			ThreadPoolJob *job;
+			ListNode *node = ListHead(&sub->outgoing);
+			job = (ThreadPoolJob*)node->item;
+			/* The new head of queue should not have already been
+			   added to the pool, else something is very wrong */
+			assert(job->jobId != STALE_JOBID);
+
+			ThreadPoolAdd(&gSendThreadPool, job, NULL);
+			job->jobId = STALE_JOBID;
+		}
+	}
+
 	if (return_code == GENA_E_NOTIFY_UNACCEPTED_REMOVE_SUB)
 		RemoveSubscriptionSID(in->sid, service);
 	free_notify_struct(in);
@@ -442,14 +446,35 @@ ExitFunction:
 	return headers;
 }
 
+void freeSubscriptionQueuedEvents(subscription *sub)
+{
+	if (ListSize(&sub->outgoing) > 0) {
+		/* The first event is discarded without dealing
+		   notify_thread_struct: there is a mirror ThreadPool entry for
+		   this one, and it will take care of the refcount etc. Other
+		   entries must be fully cleaned-up here */
+		int first = 1;
+		ListNode *node = ListHead(&sub->outgoing);
+		while (node) {
+			ThreadPoolJob *job = (ThreadPoolJob *)node->item;
+			if (first) {
+				first = 0;
+			} else {
+				free_notify_struct((notify_thread_struct *)job->arg);
+			}
+			free(node->item);
+			ListDelNode(&sub->outgoing, node, 0);
+			node = ListHead(&sub->outgoing);
+		}
+	}
+}
 
-int genaInitNotify(
+/* We take ownership of propertySet and will free it */
+static int genaInitNotifyCommon(
 	UpnpDevice_Handle device_handle,
 	char *UDN,
 	char *servId,
-	char **VarNames,
-	char **VarValues,
-	int var_count,
+	DOMString propertySet,
 	const Upnp_SID sid)
 {
 	int ret = GENA_SUCCESS;
@@ -458,19 +483,24 @@ int genaInitNotify(
 	int *reference_count = NULL;
 	char *UDN_copy = NULL;
 	char *servId_copy = NULL;
-	DOMString propertySet = NULL;
 	char *headers = NULL;
 	notify_thread_struct *thread_struct = NULL;
 
 	subscription *sub = NULL;
 	service_info *service = NULL;
 	struct Handle_Info *handle_info;
-	ThreadPoolJob job;
-
-	memset(&job, 0, sizeof(job));
+	ThreadPoolJob *job = NULL;
 
 	UpnpPrintf(UPNP_INFO, GENA, __FILE__, __LINE__,
-		"GENA BEGIN INITIAL NOTIFY");
+		"GENA BEGIN INITIAL NOTIFY COMMON");
+
+	job = (ThreadPoolJob *)malloc(sizeof(ThreadPoolJob));
+	if (job == NULL) {
+		line = __LINE__;
+		ret = UPNP_E_OUTOF_MEMORY;
+		goto ExitFunction;
+	}
+	memset(job, 0, sizeof(ThreadPoolJob));
 
 	reference_count = (int *)malloc(sizeof (int));
 	if (reference_count == NULL) {
@@ -522,21 +552,6 @@ int genaInitNotify(
 		"FOUND SUBSCRIPTION IN INIT NOTIFY: SID %s", sid);
 	sub->active = 1;
 
-	if (var_count <= 0) {
-		line = __LINE__;
-		ret = GENA_SUCCESS;
-		goto ExitFunction;
-	}
-
-	ret = GeneratePropertySet(VarNames, VarValues, var_count, &propertySet);
-	if (ret != XML_SUCCESS) {
-		line = __LINE__;
-		goto ExitFunction;
-	}
-	UpnpPrintf(UPNP_INFO, GENA, __FILE__, __LINE__,
-		"GENERATED PROPERTY SET IN INIT NOTIFY: %s",
-		propertySet);
-
 	headers = AllocGenaHeaders(propertySet);
 	if (headers == NULL) {
 		line = __LINE__;
@@ -559,28 +574,36 @@ int genaInitNotify(
 		memset(thread_struct->sid, 0, sizeof(thread_struct->sid));
 		strncpy(thread_struct->sid, sid,
 			sizeof(thread_struct->sid) - 1);
-		thread_struct->eventKey = sub->eventKey++;
+		thread_struct->ctime = time(0);
 		thread_struct->reference_count = reference_count;
 		thread_struct->device_handle = device_handle;
 
-		TPJobInit(&job, (start_routine)genaNotifyThread, thread_struct);
-		TPJobSetFreeFunction(&job, (free_routine)free_notify_struct);
-		TPJobSetPriority(&job, MED_PRIORITY);
+		TPJobInit(job, (start_routine)genaNotifyThread, thread_struct);
+		TPJobSetFreeFunction(job, (free_routine)free_notify_struct);
+		TPJobSetPriority(job, MED_PRIORITY);
 
-		ret = ThreadPoolAdd(&gSendThreadPool, &job, NULL);
+		ret = ThreadPoolAdd(&gSendThreadPool, job, NULL);
 		if (ret != 0) {
 			if (ret == EOUTOFMEM) {
 				line = __LINE__;
 				ret = UPNP_E_OUTOF_MEMORY;
 			}
 		} else {
-			line = __LINE__;
-			ret = GENA_SUCCESS;
+			ListNode *node = ListAddTail(&sub->outgoing, job);
+			if (node != NULL) {
+				((ThreadPoolJob *)node->item)->jobId = STALE_JOBID;
+				line = __LINE__;
+				ret = GENA_SUCCESS;
+			} else {
+				line = __LINE__;
+				ret = UPNP_E_OUTOF_MEMORY;
+			}
 		}
 	}
 
 ExitFunction:
-	if (ret != GENA_SUCCESS || var_count <= 0) {
+	if (ret != GENA_SUCCESS) {
+		free(job);
 		free(thread_struct);
 		free(headers);
 		ixmlFreeDOMString(propertySet);
@@ -592,12 +615,53 @@ ExitFunction:
 	HandleUnlock();
 
 	UpnpPrintf(UPNP_INFO, GENA, __FILE__, line,
-		"GENA END INITIAL NOTIFY, ret = %d",
+		"GENA END INITIAL NOTIFY COMMON, ret = %d",
 		ret);
 
 	return ret;
 }
 
+int genaInitNotify(
+	UpnpDevice_Handle device_handle,
+	char *UDN,
+	char *servId,
+	char **VarNames,
+	char **VarValues,
+	int var_count,
+	const Upnp_SID sid)
+{
+	int ret = GENA_SUCCESS;
+	int line = 0;
+	DOMString propertySet = NULL;
+
+	UpnpPrintf(UPNP_INFO, GENA, __FILE__, __LINE__,
+		"GENA BEGIN INITIAL NOTIFY");
+
+	if (var_count <= 0) {
+		line = __LINE__;
+		ret = GENA_SUCCESS;
+		goto ExitFunction;
+	}
+
+	ret = GeneratePropertySet(VarNames, VarValues, var_count, &propertySet);
+	if (ret != XML_SUCCESS) {
+		line = __LINE__;
+		goto ExitFunction;
+	}
+	UpnpPrintf(UPNP_INFO, GENA, __FILE__, __LINE__,
+		"GENERATED PROPERTY SET IN INIT NOTIFY: %s",
+		propertySet);
+
+	ret = genaInitNotifyCommon(device_handle, UDN, servId, propertySet, sid);
+
+ExitFunction:
+
+	UpnpPrintf(UPNP_INFO, GENA, __FILE__, line,
+		"GENA END INITIAL NOTIFY, ret = %d",
+		ret);
+
+	return ret;
+}
 
 int genaInitNotifyExt(
 	UpnpDevice_Handle device_handle,
@@ -609,72 +673,10 @@ int genaInitNotifyExt(
 	int ret = GENA_SUCCESS;
 	int line = 0;
 
-	int *reference_count = NULL;
-	char *UDN_copy = NULL;
-	char *servId_copy = NULL;
 	DOMString propertySet = NULL;
-	char *headers = NULL;
-	notify_thread_struct *thread_struct = NULL;
-
-	subscription *sub = NULL;
-	service_info *service = NULL;
-	struct Handle_Info *handle_info;
-	ThreadPoolJob job;
-
-	memset(&job, 0, sizeof(job));
 
 	UpnpPrintf(UPNP_INFO, GENA, __FILE__, __LINE__,
 		"GENA BEGIN INITIAL NOTIFY EXT");
-	
-	reference_count = (int *)malloc(sizeof (int));
-	if (reference_count == NULL) {
-		line = __LINE__;
-		ret = UPNP_E_OUTOF_MEMORY;
-		goto ExitFunction;
-	}
-	*reference_count = 0;
-	
-	UDN_copy = strdup(UDN);
-	if (UDN_copy == NULL) {
-		line = __LINE__;
-		ret = UPNP_E_OUTOF_MEMORY;
-		goto ExitFunction;
-	}
-
-	servId_copy = strdup(servId);
-	if( servId_copy == NULL ) {
-		line = __LINE__;
-		ret = UPNP_E_OUTOF_MEMORY;
-		goto ExitFunction;
-	}
-
-	HandleLock();
-
-	if (GetHandleInfo(device_handle, &handle_info) != HND_DEVICE) {
-		line = __LINE__;
-		ret = GENA_E_BAD_HANDLE;
-		goto ExitFunction;
-	}
-
-	service = FindServiceId(&handle_info->ServiceTable, servId, UDN);
-	if (service == NULL) {
-		line = __LINE__;
-		ret = GENA_E_BAD_SERVICE;
-		goto ExitFunction;
-	}
-	UpnpPrintf(UPNP_INFO, GENA, __FILE__, __LINE__,
-		"FOUND SERVICE IN INIT NOTFY EXT: UDN %s, ServID: %s",
-		UDN, servId);
-
-	sub = GetSubscriptionSID(sid, service);
-	if (sub == NULL || sub->active) {
-		line = __LINE__;
-		ret = GENA_E_BAD_SID;
-		goto ExitFunction;
-	}
-	UpnpPrintf( UPNP_INFO, GENA, __FILE__, __LINE__,
-		"FOUND SUBSCRIPTION IN INIT NOTIFY EXT: SID %s", sid);
-	sub->active = 1;
 
 	if (PropSet == 0) {
 		line = __LINE__;
@@ -692,59 +694,9 @@ int genaInitNotifyExt(
 		"GENERATED PROPERTY SET IN INIT EXT NOTIFY: %s",
 		propertySet);
 
-	headers = AllocGenaHeaders(propertySet);
-	if (headers == NULL) {
-		line = __LINE__;
-		ret = UPNP_E_OUTOF_MEMORY;
-		goto ExitFunction;
-	}
-
-	/* schedule thread for initial notification */
-
-	thread_struct = (notify_thread_struct *)malloc(sizeof (notify_thread_struct));
-	if (thread_struct == NULL) {
-		line = __LINE__;
-		ret = UPNP_E_OUTOF_MEMORY;
-	} else {
-		*reference_count = 1;
-		thread_struct->servId = servId_copy;
-		thread_struct->UDN = UDN_copy;
-		thread_struct->headers = headers;
-		thread_struct->propertySet = propertySet;
-		memset(thread_struct->sid, 0, sizeof(thread_struct->sid));
-		strncpy(thread_struct->sid, sid,
-			sizeof(thread_struct->sid) - 1);
-		thread_struct->eventKey = sub->eventKey++;
-		thread_struct->reference_count = reference_count;
-		thread_struct->device_handle = device_handle;
-
-		TPJobInit(&job, (start_routine)genaNotifyThread, thread_struct);
-		TPJobSetFreeFunction(&job, (free_routine)free_notify_struct);
-		TPJobSetPriority(&job, MED_PRIORITY);
-
-		ret = ThreadPoolAdd(&gSendThreadPool, &job, NULL);
-		if (ret != 0) {
-			if (ret == EOUTOFMEM) {
-				line = __LINE__;
-				ret = UPNP_E_OUTOF_MEMORY;
-			}
-		} else {
-			line = __LINE__;
-			ret = GENA_SUCCESS;
-		}
-	}
+	ret = genaInitNotifyCommon(device_handle, UDN, servId, propertySet, sid);
 
 ExitFunction:
-	if (ret != GENA_SUCCESS || PropSet == 0) {
-		free(thread_struct);
-		free(headers);
-		ixmlFreeDOMString(propertySet);
-		free(servId_copy);
-		free(UDN_copy);
-		free(reference_count);
-	}
-
-	HandleUnlock();
 
 	UpnpPrintf(UPNP_INFO, GENA, __FILE__, line,
 		"GENA END INITIAL NOTIFY EXT, ret = %d",
@@ -753,12 +705,49 @@ ExitFunction:
 	return ret;
 }
 
+/*
+ * This gets called before queuing a new event.
+ * - The list size can never go over MAX_SUBSCRIPTION_QUEUED_EVENTS so we
+ *   discard the oldest non-active event if it is already at the max
+ * - We also discard any non-active event older than MAX_SUBSCRIPTION_EVENT_AGE.
+ * non-active: any but the head of queue, which is already copied to
+ * the thread pool
+ */
+static void maybeDiscardEvents(LinkedList *listp)
+{
+	time_t now = time(0L);
 
-int genaNotifyAllExt(
+	while (ListSize(listp) > 1) {
+		ListNode *node = ListHead(listp);
+		/* The first candidate is the second event: first non-active */
+		if (node == 0 || (node = node->next) == 0) {
+			/* Major inconsistency, really, should abort here. */
+			fprintf(stderr, "gena_device: maybeDiscardEvents: "
+					"list is inconsistent\n");
+			break;
+		}
+
+		notify_thread_struct *ntsp =
+			(notify_thread_struct *)(((ThreadPoolJob *)node->item)->arg);
+		if (ListSize(listp) > g_UpnpSdkEQMaxLen ||
+			now - ntsp->ctime > g_UpnpSdkEQMaxAge) {
+			free_notify_struct(ntsp);
+			free(node->item);
+			ListDelNode(listp, node, 0);
+		} else {
+			/* If the list is smaller than the max and the oldest
+			 * task is young enough, stop pruning */
+			break;
+		}
+	}
+}
+
+/* We take ownership of propertySet and will free it */
+static int genaNotifyAllCommon(
 	UpnpDevice_Handle device_handle,
 	char *UDN,
 	char *servId,
-	IXML_Document *PropSet)
+	DOMString propertySet)
 {
 	int ret = GENA_SUCCESS;
 	int line = 0;
@@ -766,20 +755,17 @@ int genaNotifyAllExt(
 	int *reference_count = NULL;
 	char *UDN_copy = NULL;
 	char *servId_copy = NULL;
-	DOMString propertySet = NULL;
 	char *headers = NULL;
 	notify_thread_struct *thread_struct = NULL;
 
 	subscription *finger = NULL;
 	service_info *service = NULL;
 	struct Handle_Info *handle_info;
-	ThreadPoolJob job;
-
-	memset(&job, 0, sizeof(job));
 
 	UpnpPrintf(UPNP_INFO, GENA, __FILE__, __LINE__,
-		"GENA BEGIN NOTIFY ALL EXT");
+		"GENA BEGIN NOTIFY ALL COMMON");
 
+	/* Keep this allocation first */
 	reference_count = (int *)malloc(sizeof (int));
 	if (reference_count == NULL) {
 		line = __LINE__;
@@ -802,16 +788,6 @@ int genaNotifyAllExt(
 		goto ExitFunction;
 	}
 
-	propertySet = ixmlPrintNode((IXML_Node *)PropSet);
-	if (propertySet == NULL) {
-		line = __LINE__;
-		ret = UPNP_E_INVALID_PARAM;
-		goto ExitFunction;
-	}
-	UpnpPrintf(UPNP_INFO, GENA, __FILE__, __LINE__,
-		"GENERATED PROPERTY SET IN EXT NOTIFY: %s",
-		propertySet);
-
 	headers = AllocGenaHeaders(propertySet);
 	if (headers == NULL) {
 		line = __LINE__;
@@ -829,6 +805,9 @@ int genaNotifyAllExt(
 		if (service != NULL) {
 			finger = GetFirstSubscription(service);
 			while (finger) {
+				ThreadPoolJob *job = NULL;
+				ListNode *node;
+
 				thread_struct = (notify_thread_struct *)malloc(sizeof (notify_thread_struct));
 				if (thread_struct == NULL) {
 					line = __LINE__;
@@ -846,26 +825,38 @@ int genaNotifyAllExt(
 					sizeof(thread_struct->sid));
 				strncpy(thread_struct->sid, finger->sid,
 					sizeof(thread_struct->sid) - 1);
-				thread_struct->eventKey = finger->eventKey++;
+				thread_struct->ctime = time(0);
 				thread_struct->device_handle = device_handle;
-				/* if overflow, wrap to 1 */
-				if (finger->eventKey < 0) {
-					finger->eventKey = 1;
-				}
 
-				TPJobInit(&job, (start_routine)genaNotifyThread, thread_struct);
-				TPJobSetFreeFunction(&job, (free_routine)free_notify_struct);
-				TPJobSetPriority(&job, MED_PRIORITY);
-				ret = ThreadPoolAdd(&gSendThreadPool, &job, NULL);
-				if (ret != 0) {
+				maybeDiscardEvents(&finger->outgoing);
+				job = (ThreadPoolJob *)malloc(sizeof(ThreadPoolJob));
+				if (job == NULL) {
 					line = __LINE__;
-					if (ret == EOUTOFMEM) {
-						line = __LINE__;
-						ret = UPNP_E_OUTOF_MEMORY;
-					}
+					ret = UPNP_E_OUTOF_MEMORY;
 					break;
 				}
+				memset(job, 0, sizeof(ThreadPoolJob));
+				TPJobInit(job, (start_routine)genaNotifyThread, thread_struct);
+				TPJobSetFreeFunction(job, (free_routine)free_notify_struct);
+				TPJobSetPriority(job, MED_PRIORITY);
+				node = ListAddTail(&finger->outgoing, job);
 
+				/* If there is only one element on the list (which we just
+				   added), need to kickstart the threadpool */
+				if (ListSize(&finger->outgoing) == 1) {
+					ret = ThreadPoolAdd(&gSendThreadPool, job, NULL);
+					if (ret != 0) {
+						line = __LINE__;
+						if (ret == EOUTOFMEM) {
+							line = __LINE__;
+							ret = UPNP_E_OUTOF_MEMORY;
+						}
+						break;
+					}
+					if (node) {
+						((ThreadPoolJob *)(node->item))->jobId = STALE_JOBID;
+					}
+				}
 				finger = GetNextSubscription(service, finger);
 			}
 		} else {
@@ -875,7 +866,11 @@ int genaNotifyAllExt(
 	}
 
 ExitFunction:
-	if (ret != GENA_SUCCESS || *reference_count == 0) {
+	/* The only case where we want to free memory here is if the
+	   struct was never queued. Else, let the normal cleanup take place.
+	   reference_count is allocated first so it's ok to do nothing if it's 0
+	*/
+	if (reference_count && *reference_count == 0) {
 		free(headers);
 		ixmlFreeDOMString(propertySet);
 		free(servId_copy);
@@ -885,6 +880,42 @@ ExitFunction:
 
 	HandleUnlock();
 
+	UpnpPrintf(UPNP_INFO, GENA, __FILE__, line,
+		"GENA END NOTIFY ALL COMMON, ret = %d",
+		ret);
+
+	return ret;
+}
+
+
+int genaNotifyAllExt(
+	UpnpDevice_Handle device_handle,
+	char *UDN,
+	char *servId,
+	IXML_Document *PropSet)
+{
+	int ret = GENA_SUCCESS;
+	int line = 0;
+
+	DOMString propertySet = NULL;
+
+	UpnpPrintf(UPNP_INFO, GENA, __FILE__, __LINE__,
+		"GENA BEGIN NOTIFY ALL EXT");
+
+	propertySet = ixmlPrintNode((IXML_Node *)PropSet);
+	if (propertySet == NULL) {
+		line = __LINE__;
+		ret = UPNP_E_INVALID_PARAM;
+		goto ExitFunction;
+	}
+	UpnpPrintf(UPNP_INFO, GENA, __FILE__, __LINE__,
+		"GENERATED PROPERTY SET IN EXT NOTIFY: %s",
+		propertySet);
+
+	ret = genaNotifyAllCommon(device_handle, UDN, servId, propertySet);
+
+ExitFunction:
+
 	UpnpPrintf(UPNP_INFO, GENA, __FILE__, line,
 		"GENA END NOTIFY ALL EXT, ret = %d",
 		ret);
@@ -904,45 +935,11 @@ int genaNotifyAll(
 	int ret = GENA_SUCCESS;
 	int line = 0;
 
-	int *reference_count = NULL;
-	char *UDN_copy = NULL;
-	char *servId_copy = NULL;
 	DOMString propertySet = NULL;
-	char *headers = NULL;
-	notify_thread_struct *thread_struct = NULL;
-
-	subscription *finger = NULL;
-	service_info *service = NULL;
-	struct Handle_Info *handle_info;
-	ThreadPoolJob job;
-
-	memset(&job, 0, sizeof(job));
 
 	UpnpPrintf(UPNP_INFO, GENA, __FILE__, __LINE__,
 		"GENA BEGIN NOTIFY ALL");
 
-	reference_count = (int *)malloc(sizeof (int));
-	if (reference_count == NULL) {
-		line = __LINE__;
-		ret = UPNP_E_OUTOF_MEMORY;
-		goto ExitFunction;
-	}
-	*reference_count = 0;
-	
-	UDN_copy = strdup(UDN);
-	if (UDN_copy == NULL) {
-		line = __LINE__;
-		ret = UPNP_E_OUTOF_MEMORY;
-		goto ExitFunction;
-	}
-
-	servId_copy = strdup(servId);
-	if( servId_copy == NULL ) {
-		line = __LINE__;
-		ret = UPNP_E_OUTOF_MEMORY;
-		goto ExitFunction;
-	}
-
 	ret = GeneratePropertySet(VarNames, VarValues, var_count, &propertySet);
 	if (ret != XML_SUCCESS) {
 		line = __LINE__;
@@ -952,79 +949,9 @@ int genaNotifyAll(
 		"GENERATED PROPERTY SET IN EXT NOTIFY: %s",
 		propertySet);
 
-	headers = AllocGenaHeaders(propertySet);
-	if (headers == NULL) {
-		line = __LINE__;
-		ret = UPNP_E_OUTOF_MEMORY;
-		goto ExitFunction;
-	}
-
-	HandleLock();
-
-	if (GetHandleInfo(device_handle, &handle_info) != HND_DEVICE) {
-		line = __LINE__;
-		ret = GENA_E_BAD_HANDLE;
-	} else {
-		service = FindServiceId(&handle_info->ServiceTable, servId, UDN);
-		if (service != NULL) {
-			finger = GetFirstSubscription(service);
-			while (finger) {
-				thread_struct = (notify_thread_struct *)malloc(sizeof (notify_thread_struct));
-				if (thread_struct == NULL) {
-					line = __LINE__;
-					ret = UPNP_E_OUTOF_MEMORY;
-					break;
-				}
-
-				(*reference_count)++;
-				thread_struct->reference_count = reference_count;
-				thread_struct->UDN = UDN_copy;
-				thread_struct->servId = servId_copy;
-				thread_struct->headers = headers;
-				thread_struct->propertySet = propertySet;
-				memset(thread_struct->sid, 0,
-					sizeof(thread_struct->sid));
-				strncpy(thread_struct->sid, finger->sid,
-					sizeof(thread_struct->sid) - 1);
-				thread_struct->eventKey = finger->eventKey++;
-				thread_struct->device_handle = device_handle;
-				/* if overflow, wrap to 1 */
-				if (finger->eventKey < 0) {
-					finger->eventKey = 1;
-				}
-
-
-				TPJobInit(&job, (start_routine)genaNotifyThread, thread_struct);
-				TPJobSetFreeFunction(&job, (free_routine)free_notify_struct);
-				TPJobSetPriority(&job, MED_PRIORITY);
-				ret = ThreadPoolAdd(&gSendThreadPool, &job, NULL);
-				if (ret != 0) {
-					line = __LINE__;
-					if (ret == EOUTOFMEM) {
-						line = __LINE__;
-						ret = UPNP_E_OUTOF_MEMORY;
-					}
-					break;
-				}
-
-				finger = GetNextSubscription(service, finger);
-			}
-		} else {
-			line = __LINE__;
-			ret = GENA_E_BAD_SERVICE;
-		}
-	}
+	ret = genaNotifyAllCommon(device_handle, UDN, servId, propertySet);
 
 ExitFunction:
-	if (ret != GENA_SUCCESS || *reference_count == 0) {
-		free(headers);
-		ixmlFreeDOMString(propertySet);
-		free(servId_copy);
-		free(UDN_copy);
-		free(reference_count);
-	}
-
-	HandleUnlock();
 
 	UpnpPrintf(UPNP_INFO, GENA, __FILE__, line,
 		"GENA END NOTIFY ALL, ret = %d",
@@ -1113,7 +1040,7 @@ static int create_url_list(
 	/*! [out] . */
 	URL_list *out)
 {
-    size_t URLcount = 0;
+    size_t URLcount = 0, URLcount2 = 0;
     size_t i;
     int return_code = 0;
     uri_type temp;
@@ -1155,16 +1082,23 @@ static int create_url_list(
         }
         memcpy( out->URLs, URLS->buff, URLS->size );
         out->URLs[URLS->size] = 0;
-        URLcount = 0;
         for( i = 0; i < URLS->size; i++ ) {
             if( ( URLS->buff[i] == '<' ) && ( i + 1 < URLS->size ) ) {
                 if( ( ( return_code =
                         parse_uri( &out->URLs[i + 1], URLS->size - i + 1,
-                                   &out->parsedURLs[URLcount] ) ) ==
+                                   &out->parsedURLs[URLcount2] ) ) ==
                       HTTP_SUCCESS )
-                    && ( out->parsedURLs[URLcount].hostport.text.size !=
+                    && ( out->parsedURLs[URLcount2].hostport.text.size !=
                          0 ) ) {
-                    URLcount++;
+                    URLcount2++;
+                    if (URLcount2 >= URLcount)
+                        /*
+                         * break early here in case there is a bogus URL that
+                         * was skipped above. This prevents to access
+                         * out->parsedURLs[URLcount] which is beyond the
+                         * allocation.
+                         */
+                        break;
                 } else {
                     if( return_code == UPNP_E_OUTOF_MEMORY ) {
                         free( out->URLs );
@@ -1277,13 +1211,17 @@ void gena_process_subscription_request(
 		HandleUnlock();
 		goto exit_function;
 	}
-	sub->eventKey = 0;
 	sub->ToSendEventKey = 0;
 	sub->active = 0;
 	sub->next = NULL;
 	sub->DeliveryURLs.size = 0;
 	sub->DeliveryURLs.URLs = NULL;
 	sub->DeliveryURLs.parsedURLs = NULL;
+	if (ListInit(&sub->outgoing, 0, free) != 0) {
+		error_respond(info, HTTP_INTERNAL_SERVER_ERROR, request);
+		HandleUnlock();
+		goto exit_function;
+	}
 
 	/* check for valid callbacks */
 	if (httpmsg_find_hdr( request, HDR_CALLBACK, &callback_hdr) == NULL) {
