@@ -523,7 +523,12 @@ static int get_miniserver_sockets(
 	uint16_t actual_port6 = 0u;
 #endif
 	int ret_code;
+#ifdef UPNP_MINISERVER_REUSEADDR
+	int reuseaddr_on = 1;
+#else
 	int reuseaddr_on = 0;
+#endif
+
 	int sockError = UPNP_E_SUCCESS;
 	int errCode = 0;
 
@@ -583,6 +588,7 @@ static int get_miniserver_sockets(
 	memset(&__ss_v6, 0, sizeof (__ss_v6));
 	serverAddr6->sin6_family = (sa_family_t)AF_INET6;
 	inet_pton(AF_INET6, gIF_IPV6, &serverAddr6->sin6_addr);
+	serverAddr6->sin6_scope_id = gIF_INDEX;
 #endif
 	/* Getting away with implementation of re-using address:port and
 	 * instead choosing to increment port numbers.
@@ -608,9 +614,9 @@ static int get_miniserver_sockets(
 				return UPNP_E_SOCKET_BIND;
 			}
 			serverAddr4->sin_port = htons(listen_port4);
-			sockError = bind(listenfd4,
-				(struct sockaddr *)&__ss_v4,
-				sizeof (__ss_v4));
+                        sockError = bind(listenfd4,
+                                (struct sockaddr *)serverAddr4,
+                                sizeof(*serverAddr4));
 			if (sockError == SOCKET_ERROR) {
 				strerror_r(errno, errorBuffer,
 					ERROR_BUFFER_LEN);
@@ -638,9 +644,9 @@ static int get_miniserver_sockets(
 				return UPNP_E_SOCKET_BIND;
 			}
 			serverAddr6->sin6_port = htons(listen_port6);
-			sockError = bind(listenfd6,
-				(struct sockaddr *)&__ss_v6,
-				sizeof (__ss_v6));
+                        sockError = bind(listenfd6,
+                                (struct sockaddr *)serverAddr6,
+                                sizeof(*serverAddr6));
 			if (sockError == SOCKET_ERROR) {
 				strerror_r(errno, errorBuffer,
 					ERROR_BUFFER_LEN);
