@@ -129,7 +129,7 @@ static const char *gMediaTypes[] = {
 #define TEXT_INDEX        5
 
 /* general */
-#define NUM_MEDIA_TYPES       69
+#define NUM_MEDIA_TYPES       70
 #define NUM_HTTP_HEADER_NAMES 33
 
 #define ASCTIME_R_BUFFER_SIZE 26
@@ -156,6 +156,7 @@ static const char *gEncodedMediaTypes =
 	"au\0" AUDIO_STR "basic\0"
 	"avi\0" VIDEO_STR "msvideo\0"
 	"bmp\0" IMAGE_STR "bmp\0"
+	"css\0" TEXT_STR "css\0"
 	"dcr\0" APPLICATION_STR "x-director\0"
 	"dib\0" IMAGE_STR "bmp\0"
 	"dir\0" APPLICATION_STR "x-director\0"
@@ -771,7 +772,7 @@ static int GetNextRange(
  *
  * \return
  * \li \c HTTP_BAD_REQUEST
- * \li \c UPNP_E_OUTOF_MEMORY
+ * \li \c HTTP_INTERNAL_SERVER_ERROR
  * \li \c HTTP_REQUEST_RANGE_NOT_SATISFIABLE
  * \li \c HTTP_OK
  */
@@ -792,11 +793,9 @@ static int CreateHTTPRangeResponseHeader(
 	Instr->ReadSendSize = FileLength;
 	if (!ByteRangeSpecifier)
 		return HTTP_BAD_REQUEST;
-	RangeInput = malloc(strlen(ByteRangeSpecifier) + 1);
+	RangeInput = strdup(ByteRangeSpecifier);
 	if (!RangeInput)
-		return UPNP_E_OUTOF_MEMORY;
-	memset(RangeInput, 0, strlen(ByteRangeSpecifier) + 1);
-	strncpy(RangeInput, ByteRangeSpecifier, strlen(ByteRangeSpecifier));
+		return HTTP_INTERNAL_SERVER_ERROR;
 	/* CONTENT-RANGE: bytes 222-3333/4000  HTTP_PARTIAL_CONTENT */
 	if (StrStr(RangeInput, "bytes") == NULL ||
 	    (Ptr = StrStr(RangeInput, "=")) == NULL) {
@@ -830,7 +829,7 @@ static int CreateHTTPRangeResponseHeader(
 				(int64_t)FileLength);
 			if (rc < 0 || (unsigned int) rc >= sizeof(Instr->RangeHeader)) {
 				free(RangeInput);
-				return UPNP_E_OUTOF_MEMORY;
+				return HTTP_INTERNAL_SERVER_ERROR;
 			}
 		} else if (FirstByte >= 0 && LastByte == -1
 			   && FirstByte < FileLength) {
@@ -845,7 +844,7 @@ static int CreateHTTPRangeResponseHeader(
 				(int64_t)FileLength);
 			if (rc < 0 || (unsigned int) rc >= sizeof(Instr->RangeHeader)) {
 				free(RangeInput);
-				return UPNP_E_OUTOF_MEMORY;
+				return HTTP_INTERNAL_SERVER_ERROR;
 			}
 		} else if (FirstByte == -1 && LastByte > 0) {
 			if (LastByte >= FileLength) {
@@ -864,13 +863,13 @@ static int CreateHTTPRangeResponseHeader(
 					sizeof(Instr->RangeHeader),
 					"CONTENT-RANGE: bytes %" PRId64
 					"-%" PRId64 "/%" PRId64 "\r\n",
-					(int64_t)(FileLength - LastByte + 1),
-					(int64_t)FileLength,
+					(int64_t)(FileLength - LastByte),
+					(int64_t)FileLength - 1,
 					(int64_t)FileLength);
 			}
 			if (rc < 0 || (unsigned int) rc >= sizeof(Instr->RangeHeader)) {
 				free(RangeInput);
-				return UPNP_E_OUTOF_MEMORY;
+				return HTTP_INTERNAL_SERVER_ERROR;
 			}
 		} else {
 			free(RangeInput);
@@ -891,7 +890,7 @@ static int CreateHTTPRangeResponseHeader(
  *
  * \return
  * \li \c HTTP_BAD_REQUEST
- * \li \c UPNP_E_OUTOF_MEMORY
+ * \li \c HTTP_INTERNAL_SERVER_ERROR
  * \li \c HTTP_REQUEST_RANGE_NOT_SATISFIABLE
  * \li \c HTTP_OK
  */
@@ -912,7 +911,7 @@ static int CheckOtherHTTPHeaders(
 
 	TmpBuf = (char *)malloc(TmpBufSize);
 	if (!TmpBuf)
-		return UPNP_E_OUTOF_MEMORY;
+		return HTTP_INTERNAL_SERVER_ERROR;
 	node = ListHead(&Req->headers);
 	while (node != NULL) {
 		header = (http_header_t *) node->item;
@@ -925,7 +924,7 @@ static int CheckOtherHTTPHeaders(
 			TmpBufSize = header->value.length + 1;
 			TmpBuf = (char *)malloc(TmpBufSize);
 			if (!TmpBuf)
-				return UPNP_E_OUTOF_MEMORY;
+				return HTTP_INTERNAL_SERVER_ERROR;
 		}
 		memcpy(TmpBuf, header->value.buf, header->value.length);
 		TmpBuf[header->value.length] = '\0';
@@ -1009,13 +1008,91 @@ static int CheckOtherHTTPHeaders(
 	return RetCode;
 }
 
+/*!
+ * \brief Build an array of unrecognized headers.
+ *
+ * \return nothing
+ */
+#define MAX_EXTRA_HEADERS 128
+static int ExtraHTTPHeaders(
+	/*! [in] HTTP Request message. */
+	http_message_t *Req,
+	struct Extra_Headers **ExtraHeaders)
+{
+	http_header_t *header;
+	ListNode *node;
+	int index, nb_extra = 0;
+	struct Extra_Headers *extra_headers;
+
+	node = ListHead(&Req->headers);
+	extra_headers = *ExtraHeaders =
+		(struct Extra_Headers*) malloc(MAX_EXTRA_HEADERS * sizeof(struct Extra_Headers));
+	if (!extra_headers) {
+		return HTTP_INTERNAL_SERVER_ERROR;
+	}
+	while (node != NULL) {
+		header = (http_header_t *) node->item;
+		/* find header type. */
+		index = map_str_to_int((const char *)header->name.buf,
+				header->name.length, Http_Header_Names,
+				NUM_HTTP_HEADER_NAMES, FALSE);
+		if (index < 0) {
+			extra_headers->name = (char *)malloc(header->name.length + 1);
+			extra_headers->value = (char *)malloc(header->value.length + 1);
+			if (!extra_headers->name || !extra_headers->value) {
+				/* cleanup will be made by caller */
+				return HTTP_INTERNAL_SERVER_ERROR;
+			}
+			memcpy(extra_headers->name, header->name.buf, header->name.length);
+			memcpy(extra_headers->value, header->value.buf, header->value.length);
+			extra_headers->name[header->name.length] = '\0';
+			extra_headers->value[header->value.length] = '\0';
+			extra_headers->resp = NULL;
+
+			extra_headers++;
+			nb_extra++;
+
+			if (nb_extra == MAX_EXTRA_HEADERS - 1) {
+				break;
+			}
+		}
+		node = ListNext(&Req->headers, node);
+	}
+	extra_headers->name = extra_headers->value = extra_headers->resp = NULL;
+	return HTTP_OK;
+}
+
+static void FreeExtraHTTPHeaders(
+	/*! [in] extra HTTP headers to free. */
+	struct Extra_Headers *ExtraHeaders)
+{
+	struct Extra_Headers *extra_headers = ExtraHeaders;
+
+	if (!ExtraHeaders) {
+		return;
+	}
+
+	while (extra_headers->name) {
+		free(extra_headers->name);
+		if (extra_headers->value) free(extra_headers->value);
+		if (extra_headers->resp) ixmlFreeDOMString(extra_headers->resp);
+		extra_headers++;
+	}
+
+	free(ExtraHeaders);
+}
+
+
 /*!
  * \brief Processes the request and returns the result in the output parameters.
  *
  * \return
  * \li \c HTTP_BAD_REQUEST
- * \li \c UPNP_E_OUTOF_MEMORY
+ * \li \c HTTP_INTERNAL_SERVER_ERROR
  * \li \c HTTP_REQUEST_RANGE_NOT_SATISFIABLE
+ * \li \c HTTP_FORBIDDEN
+ * \li \c HTTP_NOT_FOUND
+ * \li \c HTTP_NOT_ACCEPTABLE
  * \li \c HTTP_OK
  */
 static int process_request(
@@ -1045,7 +1122,7 @@ static int process_request(
 	int resp_minor;
 	int alias_grabbed;
 	size_t dummy;
-	const char *extra_headers = NULL;
+	struct Extra_Headers *extra_headers = NULL;
 
 	print_http_headers(req);
 	url = &req->uri;
@@ -1109,7 +1186,12 @@ static int process_request(
 	}
 	if (using_virtual_dir) {
 		if (req->method != HTTPMETHOD_POST) {
+			if ((code = ExtraHTTPHeaders(req, &extra_headers)) != HTTP_OK) {
+				err_code = code;
+				goto error_handler;
+			}
 			/* get file info */
+			finfo.extra_headers = extra_headers;
 			if (virtualDirCallback.
 			    get_info(filename->buf, &finfo) != 0) {
 				err_code = HTTP_NOT_FOUND;
@@ -1202,20 +1284,18 @@ static int process_request(
 	}
 	RespInstr->ReadSendSize = finfo.file_length;
 	/* Check other header field. */
-	if ((err_code =
-	     CheckOtherHTTPHeaders(req, RespInstr,
+	if ((code =
+		 CheckOtherHTTPHeaders(req, RespInstr,
 				   finfo.file_length)) != HTTP_OK) {
+		err_code = code;
 		goto error_handler;
 	}
+
 	if (req->method == HTTPMETHOD_POST) {
 		*rtype = RESP_POST;
-		err_code = UPNP_E_SUCCESS;
+		err_code = HTTP_OK;
 		goto error_handler;
 	}
-	/*extra_headers = UpnpFileInfo_get_ExtraHeaders(finfo); */
-	if (!extra_headers) {
-		extra_headers = "";
-	}
 
 	/* Check if chunked encoding should be used. */
 	if (using_virtual_dir && finfo.file_length == UPNP_USING_CHUNKED) {
@@ -1235,7 +1315,7 @@ static int process_request(
 		/* Content-Range: bytes 222-3333/4000  HTTP_PARTIAL_CONTENT */
 		/* Transfer-Encoding: chunked */
 		if (http_MakeMessage(headers, resp_major, resp_minor,
-		    "R" "T" "GKLD" "s" "tcS" "Xc" "sCc",
+		    "R" "T" "GKLD" "s" "tcS" "Xc" "ECc",
 		    HTTP_PARTIAL_CONTENT,	/* status code */
 		    finfo.content_type,	/* content type */
 		    RespInstr,	/* range info */
@@ -1247,61 +1327,55 @@ static int process_request(
 		}
 	} else if (RespInstr->IsRangeActive && !RespInstr->IsChunkActive) {
 		/* Content-Range: bytes 222-3333/4000  HTTP_PARTIAL_CONTENT */
-		/* Transfer-Encoding: chunked */
 		if (http_MakeMessage(headers, resp_major, resp_minor,
-		    "R" "N" "T" "GLD" "s" "tcS" "Xc" "sCc",
-		    HTTP_PARTIAL_CONTENT,	/* status code */
-		    RespInstr->ReadSendSize,	/* content length */
-		    finfo.content_type,	/* content type */
-		    RespInstr,	/* range info */
-		    RespInstr,	/* language info */
-		    "LAST-MODIFIED: ",
-		    &finfo.last_modified,
-		    X_USER_AGENT, extra_headers) != 0) {
+			"R" "N" "T" "GLD" "s" "tcS" "Xc" "ECc",
+			HTTP_PARTIAL_CONTENT,	/* status code */
+			RespInstr->ReadSendSize,	/* content length */
+			finfo.content_type,	/* content type */
+			RespInstr,	/* range info */
+			RespInstr,	/* language info */
+			"LAST-MODIFIED: ",
+			&finfo.last_modified,
+			X_USER_AGENT, extra_headers) != 0) {
 			goto error_handler;
 		}
 	} else if (!RespInstr->IsRangeActive && RespInstr->IsChunkActive) {
-		/* Content-Range: bytes 222-3333/4000  HTTP_PARTIAL_CONTENT */
 		/* Transfer-Encoding: chunked */
 		if (http_MakeMessage(headers, resp_major, resp_minor,
-		    "RK" "TLD" "s" "tcS" "Xc" "sCc",
-		    HTTP_OK,	/* status code */
-		    finfo.content_type,	/* content type */
-		    RespInstr,	/* language info */
-		    "LAST-MODIFIED: ",
-		    &finfo.last_modified,
-		    X_USER_AGENT, extra_headers) != 0) {
+			"RK" "TLD" "s" "tcS" "Xc" "ECc",
+			HTTP_OK,	/* status code */
+			finfo.content_type,	/* content type */
+			RespInstr,	/* language info */
+			"LAST-MODIFIED: ",
+			&finfo.last_modified,
+			X_USER_AGENT, extra_headers) != 0) {
 			goto error_handler;
 		}
 	} else {
 		/* !RespInstr->IsRangeActive && !RespInstr->IsChunkActive */
 		if (RespInstr->ReadSendSize >= 0) {
-			/* Content-Range: bytes 222-3333/4000  HTTP_PARTIAL_CONTENT */
-			/* Transfer-Encoding: chunked */
 			if (http_MakeMessage(headers, resp_major, resp_minor,
-			    "R" "N" "TLD" "s" "tcS" "Xc" "sCc",
-			    HTTP_OK,	/* status code */
-			    RespInstr->ReadSendSize,	/* content length */
-			    finfo.content_type,	/* content type */
-			    RespInstr,	/* language info */
-			    "LAST-MODIFIED: ",
-			    &finfo.last_modified,
-			    X_USER_AGENT,
-			    extra_headers) != 0) {
+				"R" "N" "TLD" "s" "tcS" "Xc" "ECc",
+				HTTP_OK,	/* status code */
+				RespInstr->ReadSendSize,	/* content length */
+				finfo.content_type,	/* content type */
+				RespInstr,	/* language info */
+				"LAST-MODIFIED: ",
+				&finfo.last_modified,
+				X_USER_AGENT,
+				extra_headers) != 0) {
 				goto error_handler;
 			}
 		} else {
-			/* Content-Range: bytes 222-3333/4000  HTTP_PARTIAL_CONTENT */
-			/* Transfer-Encoding: chunked */
 			if (http_MakeMessage(headers, resp_major, resp_minor,
-			    "R" "TLD" "s" "tcS" "Xc" "sCc",
-			    HTTP_OK,	/* status code */
-			    finfo.content_type,	/* content type */
-			    RespInstr,	/* language info */
-			    "LAST-MODIFIED: ",
-			    &finfo.last_modified,
-			    X_USER_AGENT,
-			    extra_headers) != 0) {
+				"R" "TLD" "s" "tcS" "Xc" "ECc",
+				HTTP_OK,	/* status code */
+				finfo.content_type,	/* content type */
+				RespInstr,	/* language info */
+				"LAST-MODIFIED: ",
+				&finfo.last_modified,
+				X_USER_AGENT,
+				extra_headers) != 0) {
 				goto error_handler;
 			}
 		}
@@ -1322,12 +1396,13 @@ static int process_request(
 	if (req->method == HTTPMETHOD_SIMPLEGET) {
 		membuffer_destroy(headers);
 	}
-	err_code = UPNP_E_SUCCESS;
+	err_code = HTTP_OK;
 
  error_handler:
 	free(request_doc);
+	FreeExtraHTTPHeaders(extra_headers);
 	ixmlFreeDOMString(finfo.content_type);
-	if (err_code != UPNP_E_SUCCESS && alias_grabbed) {
+	if (err_code != HTTP_OK && alias_grabbed) {
 		alias_release(alias);
 	}
 
@@ -1340,7 +1415,8 @@ static int process_request(
  * \return
  * \li \c HTTP_INTERNAL_SERVER_ERROR
  * \li \c HTTP_UNAUTHORIZED
- * \li \c HTTP_REQUEST_RANGE_NOT_SATISFIABLE
+ * \li \c HTTP_BAD_REQUEST
+ * \li \c HTTP_SERVICE_UNAVAILABLE
  * \li \c HTTP_OK
  */
 static int http_RecvPostMessage(
@@ -1369,9 +1445,13 @@ static int http_RecvPostMessage(
 		if (Fp == NULL)
 			return HTTP_INTERNAL_SERVER_ERROR;
 	} else {
+#ifdef UPNP_ENABLE_POST_WRITE
 		Fp = fopen(filename, "wb");
 		if (Fp == NULL)
 			return HTTP_UNAUTHORIZED;
+#else
+		return HTTP_NOT_FOUND;
+#endif
 	}
 	parser->position = POS_ENTITY;
 	do {
@@ -1426,7 +1506,7 @@ static int http_RecvPostMessage(
 					goto ExitFunction;
 				}
 			} else {
-				ret_code = num_read;
+				ret_code = HTTP_SERVICE_UNAVAILABLE;
 				goto ExitFunction;
 			}
 		}
@@ -1489,7 +1569,7 @@ void web_server_callback(http_parser_t *parser, INOUT http_message_t *req,
 	/*the type of request. */
 	ret = process_request(req, &rtype, &headers, &filename, &xmldoc,
 		&RespInstr);
-	if (ret != UPNP_E_SUCCESS) {
+	if (ret != HTTP_OK) {
 		/* send error code */
 		http_SendStatusResponse(info, ret, req->major_version,
 			req->minor_version);
@@ -1531,7 +1611,7 @@ void web_server_callback(http_parser_t *parser, INOUT http_message_t *req,
 			/* Send response. */
 			http_MakeMessage(&headers, 1, 1,
 				"RTLSXcCc",
-				ret, "text/html", X_USER_AGENT);
+				ret, "text/html", &RespInstr, X_USER_AGENT);
 			http_SendMessage(info, &timeout, "b",
 				headers.buf, headers.length);
 			break;
