@@ -49,6 +49,12 @@
 extern "C" {
 #endif
 
+#ifdef WIN32
+struct tm *http_gmtime_r(const time_t *clock, struct tm *result);
+#else
+#define http_gmtime_r gmtime_r
+#endif
+
 int http_CancelHttpGet(IN void *Handle);
 
 /*!
